import unittest

from PyShift.test.base_test import BaseTest

class TestSnippet(BaseTest):    
    def testIt(self): 
        print("Python code here")

        
if __name__ == '__main__':
    unittest.main()